package com.miya.draw.manager;

import cn.hutool.core.bean.copier.BeanCopier;
import cn.hutool.core.bean.copier.CopyOptions;
import cn.hutool.core.date.LocalDateTimeUtil;
import com.miya.draw.config.business.BusinessConfig;
import com.miya.draw.config.constant.ConStants;
import com.miya.draw.config.exception.BusinessException;
import com.miya.draw.entity.Account;
import com.miya.draw.entity.AccountPrize;
import com.miya.draw.entity.Prize;
import com.miya.draw.entity.SyncTrade;
import com.miya.draw.enums.ErrorEnum;
import com.miya.draw.enums.PrizeTypeEnum;
import com.miya.draw.enums.TradeTypeEnum;
import com.miya.draw.enums.YesOrNoEnum;
import com.miya.draw.model.DrawModel;
import com.miya.draw.request.DrawUserInfoGainRequest;
import com.miya.draw.service.*;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.redisson.api.RLock;
import org.redisson.api.RedissonClient;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.time.LocalDateTime;
import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

/**
 * 抽奖Manager
 *
 * @author jack
 */
@Component
@Slf4j
public class DrawManagerImpl implements IDrawManager {

    @Autowired
    private IAccountService accountService;

    @Autowired
    private BusinessConfig businessConfig;

    @Autowired
    private IAccountPrizeService accountPrizeService;

    @Autowired
    private ISyncTradeService syncTradeService;

    @Autowired
    private IActivityStoreService activityStoreService;

    @Autowired
    private IActivityGoodsService activityGoodsService;

    @Autowired
    private RedissonClient redissonClient;

    @Autowired
    private IPrizeService prizeService;

    @Autowired
    private IOtherManager otherManager;

    @Override
    public DrawModel gainActivityUserInfo(DrawUserInfoGainRequest request) {
        try {
            Account account = new Account();
            account.setName(StringUtils.isNotBlank(request.getUserName()) ? request.getUserName() : request.getPhone());
            account.setPhone(request.getPhone());
            account.setToken(request.getToken());
            accountService.saveOrUpdateNotById(account);
        } catch (Exception e) {
            log.error("保存用户信息出错:{}", e.getMessage());
        }

        LocalDateTime nowTime = LocalDateTime.now();
        LocalDateTime start = nowTime.minusDays(3);
        LocalDateTime end = LocalDateTimeUtil.endOfDay(nowTime);
        Integer maxDrawCountLimit = businessConfig.getDrawMaxCount();
        Integer alreadyDrawCount = accountPrizeService.gainAlreadyDrawCount(request.getPhone(), start, end);
        List<String> storeIdList = activityStoreService.listAllActivityStoreId();
        List<String> goodCodeList = activityGoodsService.listAllActivityGoods();
        List<SyncTrade> buyGoodList = syncTradeService.listCanDrawGoods(request.getPhone(), start, end, storeIdList, goodCodeList);
        BigDecimal buyTotal = gainBuyTotal(buyGoodList);
        Long purchasedAmount = buyTotal.multiply(BigDecimal.valueOf(100)).longValue();
        Integer drawCountTotal = buyTotal.divide(new BigDecimal(businessConfig.getDrawCountUnit()), RoundingMode.DOWN).intValue();
        drawCountTotal = Math.max(drawCountTotal - alreadyDrawCount, 0);
        drawCountTotal = Math.min(businessConfig.getDrawMaxCount(), drawCountTotal);
        Integer joinActivityPeopleNum = accountPrizeService.countJoinActivityPeople();
        return DrawModel.builder()
                .alreadyDrawCount(alreadyDrawCount)
                .drawCountTotal(drawCountTotal)
                .maxDrawCountLimit(maxDrawCountLimit)
                .numOfDrawCount(Math.max(0, maxDrawCountLimit - drawCountTotal))
                .purchasedAmount(purchasedAmount)
                .taxAmountCache(Math.max(0L, BigDecimal.valueOf(businessConfig.getDrawMaxMoney()).subtract(buyTotal).multiply(BigDecimal.valueOf(100)).longValue()))
                .numOfJoinActivityPeople(joinActivityPeopleNum)
                .build();
    }

    /**
     * 获取购买的交易总额
     *
     * @param buyGoodList 购买商品列表
     * @return 购买总金额
     */
    private BigDecimal gainBuyTotal(List<SyncTrade> buyGoodList) {
        BigDecimal buyTotal = buyGoodList.stream().filter(x -> TradeTypeEnum.SALE.getType().equalsIgnoreCase(x.getType())).map(syncTrade -> syncTrade.getStdAmount().subtract(syncTrade.getFavAmount()))
                .reduce(BigDecimal.ZERO, BigDecimal::add);
        BigDecimal refundTotal = buyGoodList.stream().filter(x -> TradeTypeEnum.REFUND.getType().equalsIgnoreCase(x.getType())).map(syncTrade -> syncTrade.getStdAmount().subtract(syncTrade.getFavAmount()))
                .reduce(BigDecimal.ZERO, BigDecimal::add);
        buyTotal = buyTotal.compareTo(refundTotal) >= 0 ? buyTotal.subtract(refundTotal) : BigDecimal.ZERO;
        return buyTotal;
    }

    @Override
    public List<DrawModel> listMinePrizeList(String phone) {
        List<AccountPrize> prizeList = accountPrizeService.listMinePrize(phone);
        return getDrawModels(prizeList);
    }

    @Override
    public void updateMinePrizeAddress(String phone, Long myPrizeId, String consigneeName, String consigneePhone, String consigneeAddress, String consigneeIdcard) {
        AccountPrize accountPrize = accountPrizeService.getById(myPrizeId);
        // 有奖品 手机号正确 且信息没填
        if (!Objects.isNull(accountPrize) && phone.equals(accountPrize.getAccountPhone()) && PrizeTypeEnum.PHYSICAL.getCode().equals(accountPrize.getPrizeType())) {
            if (StringUtils.isBlank(accountPrize.getConsigneeName()) && StringUtils.isBlank(accountPrize.getConsigneePhone())
                    && StringUtils.isBlank(accountPrize.getConsigneeAddress()) && StringUtils.isBlank(accountPrize.getConsigneeIdcard())) {
                AccountPrize prizeUpdate = new AccountPrize();
                prizeUpdate.setId(myPrizeId);
                prizeUpdate.setConsigneeAddress(consigneeAddress);
                prizeUpdate.setConsigneeIdcard(consigneeIdcard);
                prizeUpdate.setConsigneeName(consigneeName);
                prizeUpdate.setConsigneePhone(consigneePhone);
                // 填写信息后重新同步
                prizeUpdate.setSyncStatus(YesOrNoEnum.NO.getCode());
                accountPrizeService.saveOrUpdate(prizeUpdate);
            } else {
                throw new BusinessException(ErrorEnum.PARAM_CHECK_EXP, "收货信息已填写!");
            }
        }
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public DrawModel drawClick(String phone) throws InterruptedException {
        // 同一用户不可多次重复点击
        RLock lock = redissonClient.getLock(buildKey(phone));
        try {
            lock.tryLock(5, 10, TimeUnit.SECONDS);
            // 1 验证是否可以抽奖
            LocalDateTime nowTime = LocalDateTime.now();
            LocalDateTime start = nowTime.minusDays(3);
            LocalDateTime end = LocalDateTimeUtil.endOfDay(nowTime);
            List<SyncTrade> buyGoodList = syncTradeService.listCanDrawGoods(phone, start, end, activityStoreService.listAllActivityStoreId(), activityGoodsService.listAllActivityGoods());
            BigDecimal buyTotal = gainBuyTotal(buyGoodList);
            int drawCountTotal = buyTotal.divide(new BigDecimal(businessConfig.getDrawCountUnit()), RoundingMode.DOWN).intValue();
            drawCountTotal = Math.min(businessConfig.getDrawMaxCount(), drawCountTotal);
            if (drawCountTotal <= 0) {
                throw new BusinessException(ErrorEnum.DRAW_CHECK_EXP, "抱歉，你的购物金额不足，已无抽奖次数");
            }
            Integer alreadyDrawCount = accountPrizeService.gainAlreadyDrawCount(phone, start, end);
            // 可抽奖次数> 已抽奖次数
            Prize prize;
            AccountPrize accountPrize;
            if (drawCountTotal - alreadyDrawCount > 0) {
                // 获得奖品
                List<Prize> prizeList = prizeService.listAllPrizes();
                if (CollectionUtils.isEmpty(prizeList)) {
                    throw new BusinessException(ErrorEnum.DRAW_CHECK_EXP, "活动已结束!");
                }
                // 抽奖
                prize = selectCoupon(prizeList);
                // 乐观锁效验抽到的奖品它合不合规
                if (Integer.valueOf(1).equals(prizeService.updateCount(prize))) {
                    log.info("phone={}的用户抽中了={}", phone, prize);
                } else {
                    // 测试update=0的是否也会回滚
                    prize = prizeList.stream().filter(x -> PrizeTypeEnum.VIRTUAL.getCode().equals(x.getPrizeType())).max(Comparator.comparing(Prize::getId))
                            .orElseThrow(() -> new BusinessException(ErrorEnum.DRAW_CHECK_EXP, "当前抽奖人数过多，请稍候重试!"));
                }
                // 记录抽奖信息
                accountPrize = new AccountPrize();
                accountPrize.setPrizeId(prize.getId());
                accountPrize.setAccountName(phone);
                accountPrize.setAccountPhone(phone);
                accountPrize.setPrizeName(prize.getPrizeName());
                accountPrize.setPrizeType(prize.getPrizeType());
                accountPrizeService.saveOrUpdate(accountPrize);
                // 发券
                if (PrizeTypeEnum.VIRTUAL.getCode().equals(prize.getPrizeType()) && StringUtils.isNotBlank(prize.getCouponId())) {
                    otherManager.sendCoupon(prize.getCouponId(), phone, accountPrize.getId());
                }
            } else {
                if (alreadyDrawCount >= businessConfig.getDrawMaxCount()) {
                    throw new BusinessException(ErrorEnum.DRAW_CHECK_EXP, "抱歉,你的今日可抽奖次数已用完,请明日再来哦");
                } else {
                    throw new BusinessException(ErrorEnum.DRAW_CHECK_EXP, "抱歉，你的购物金额不足，已无抽奖次数");
                }
            }
            // 2 奖品
            BeanCopier<DrawModel> beanCopier = BeanCopier.create(prize, DrawModel.builder().build(), CopyOptions.create());
            DrawModel drawModel = beanCopier.copy();
            drawModel.setMyPrizeId(accountPrize.getId());
            drawModel.setId(prize.getId());
            return drawModel;
        } catch (Exception e) {
            log.error("抽奖异常:{}", e.getMessage());
            throw e;
        } finally {
            lock.unlock();
        }
    }

    @Override
    public List<DrawModel> listAllWinningInfo(Integer size) {
        List<AccountPrize> prizeList = accountPrizeService.listAllWinningInfo(size);
        return getDrawModels(prizeList);
    }

    /**
     * 统一封装
     *
     * @param prizeList
     * @return
     */
    private List<DrawModel> getDrawModels(List<AccountPrize> prizeList) {
        List<DrawModel> drawModelList = new ArrayList<>();
        if (CollectionUtils.isNotEmpty(prizeList)) {
            drawModelList = prizeList.stream().map(accountPrize -> {
                BeanCopier<DrawModel> copier = BeanCopier.create(accountPrize, DrawModel.builder().build(), CopyOptions.create());
                return copier.copy();
            }).collect(Collectors.toList());
        }

        return drawModelList;
    }

    /**
     * 抽奖
     *
     * @param prizeList
     * @return
     */
    private Prize selectCoupon(List<Prize> prizeList) {
        TreeMap<Double, Prize> weightMap = new TreeMap<>();
        for (Prize prize : prizeList) {
            //统一转为double
            double lastWeight = weightMap.size() == 0 ? 0 : weightMap.lastKey();
            //权重累加
            weightMap.put(prize.getPrizeOdds().doubleValue() + lastWeight, prize);
        }
        double randomWeight = weightMap.lastKey() * Math.random();
        SortedMap<Double, Prize> tailMap = weightMap.tailMap(randomWeight, false);
        return weightMap.get(tailMap.firstKey());
    }

    /**
     * 加锁可以
     *
     * @param phone
     * @return
     */
    private String buildKey(String phone) {
        StringBuffer sb = new StringBuffer();
        return sb.append(ConStants.SHIZU_DRAW_PREFIX).append(phone).toString().intern();
    }
}
